/*
*      @(#) SRC/common/vtools $Id: hdranal.c,v 1.2 2003/09/22 09:23:18 ajosey Exp $
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* The Hdranal tool is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; see the file COPYING.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

char	*copyright[] = {
		"(C) Copyright 1991 X/Open Company Limited"
};

/***********************************************************************

NAME:		hdranal.c - main entry point for hdranal(vtools)

PROJECT:	VSX (X/OPEN Validation Suite)
MODULE CONTAINS:
		main()
SYNOPSIS:
		cc -E [options] temp.c | hdranal [debug_on]

DESCRIPTION:
	Hdranal uses a stripped-down version of the GNU C compiler to
	identify symbols made visible by system headers.  To use the
	program a file containing the line "#include <header_name.h>"
	(preceded by <sys/types.h> if required) is first created.
	This is preprocessed using "cc -E" and the result fed to the
	standard input of hdranal.

	The name of the header is supplied as the first argument to
	hdranal.  This is used in the output file names with ".A",
	".M" and ".P" appended.  The ".A" and ".M" files are used in
	namespace pollution header tests.  The ".P" file is used in
	tests for existence of prototypes.
	
	The ".A" file contains a line of the form
		chkdecl("name");
	for each declared or 'typedef'ed symbol name, and
		chktag("name");
	for each struct/union/enum tag.

	The ".M" file contains a line of the form
		#define name "name unprotected"<garbage-string>
	for each struct/union member name.

	For these two files, symbol names that are in a reserved class
	of names for the given header are ignored.

	The ".P" file contains a line of the form
		#define name_HAS_PROTOTYPE
	for each function which is declared with a prototype.

	If hdranal is passed more than one argument it turns on yacc
	debug mode.

	Hdranal exits with code 2 if no arguments are supplied or code 1
	if file handling errors occur.  Exit code 1 is also given if a
	syntax error occurs (indicated by a non-zero return from yyparse())
	as this means some symbols may not have been identified.  Other C
	language errors reported are for information only and do not
	cause a non-zero exit code.  

RELATED DOCUMENTS:
	hdranal(vtools) - full manual entry

AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	26 Jan 1990
MODIFICATIONS:
		Geoff Clare, 18 May 1990
			Disable #defines in '.M' file in XPG3 mode

		Geoff Clare, 27 Aug 1992
			Align with interpretation VWG/009/063092.

		Geoff Clare, 20 June 1994
			Removed redundant 3rd argument to error_with_decl().
			Changed warn_decl() args to correct types.
			Define warning() in prototype form.

		Geoff Clare, 23 June 1999
			Deleted bcopy(), to avoid conflict with standard
			function.

		Geoff Clare, December 2001
			Changes for XSH6/POSIX01.

************************************************************************/

#include <std.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "tree.h"
#include <sysdep.h>

extern int	yyparse(void);
extern int	yydebug;

public	int	pedantic = 0;
public	FILE	*Afp;
public	FILE	*Mfp;
public	FILE	*Pfp;
public	char	*headerfile;	/* for hdrallow() */

private void	openfiles(char *);

private char	Afile[20];
private char	Mfile[20];
private char	Pfile[20];

int
main(argc, argv)
int argc;
char **argv;
{
	set_pname(argv[0]);

	if (argc <= 1)
	{
		(void) fprintf(stderr, "usage: %s headerfile [debugon]\n",
			get_pname());
		exit(2);
	}
	headerfile = argv[1];

	if (argc > 2)
		yydebug = 1;

	openfiles(headerfile);

	init_tree();
	init_lex();
	init_decl_processing();

	(void) ungetc(check_newline(), stdin);

	if (yyparse() != 0)
	{
		vsxerror("could not parse input correctly (yyparse() != 0)");
		exit(FAIL);
	}

	(void) fflush(Afp);
	(void) fflush(Mfp);
	(void) fflush(Pfp);
	if (ferror(Afp) || ferror(Mfp) || ferror(Pfp))
	{
		if (ferror(Afp))
			vsxerror("write error on \"%s\"", Afile);
		if (ferror(Mfp))
			vsxerror("write error on \"%s\"", Mfile);
		if (ferror(Pfp))
			vsxerror("write error on \"%s\"", Pfile);
		exit(FAIL);
	}

	exit(SUCCEED);
	/* NOTREACHED */
}

private void
openfiles(hdrfile)
char *hdrfile;
{
	/* Open the ".A" and ".M" output files */

	char *comment = "/* this comment prevents empty include files */\n";

	if (strlen(hdrfile) > sizeof(Afile) - 3)
	{
		vsxerror("name too long: \"%s\"", hdrfile);
		exit(FAIL);
	}

	(void) strcpy(Afile, hdrfile);
	(void) strcpy(Mfile, hdrfile);
	(void) strcpy(Pfile, hdrfile);
	(void) strcat(Afile, ".A");
	(void) strcat(Mfile, ".M");
	(void) strcat(Pfile, ".P");

	if ((Afp = fopen(Afile, "w")) == NULL)
	{
		vsxerror("cannot open \"%s\" for writing", Afile);
		exit(FAIL);
	}
	(void) fputs(comment, Afp);

	if ((Mfp = fopen(Mfile, "w")) == NULL)
	{
		vsxerror("cannot open \"%s\" for writing", Mfile);
		exit(FAIL);
	}
	(void) fputs(comment, Mfp);

	if ((Pfp = fopen(Pfile, "w")) == NULL)
	{
		vsxerror("cannot open \"%s\" for writing", Pfile);
		exit(FAIL);
	}
	(void) fputs(comment, Pfp);
}

/* ARGSUSED */
void
decl_rest_of_compilation (decl, asmspec, top_level, at_end)
tree decl;
tree asmspec;
int top_level;
int at_end;
{
	if (top_level)
	{
		char *name = IDENTIFIER_POINTER (DECL_NAME (decl));
		if (!hdrallow(name, 0))
      			(void) fprintf(Afp, "\tchkdecl(\"%s\");\n", name);
	}
}

void
rest_of_compilation (decl)
tree decl;
{
	char *name = IDENTIFIER_POINTER (DECL_NAME (decl));
	if (!hdrallow(name, 0))
		(void) fprintf(Afp, "\tchkdecl(\"%s\");\n", name);
}

void
error(fmt, arg)
char *fmt, *arg;
{
	(void) fprintf(stderr, "ERROR: ");
	(void) fprintf(stderr, fmt, arg);
	(void) fprintf(stderr, "\n");
}

void
error_with_decl(decl, s)
tree decl;
char *s;
{
	char *name;

	if (DECL_NAME(decl))
		name = IDENTIFIER_POINTER(DECL_NAME(decl));
	else
		name = "((anonymous))";

	(void) fprintf(stderr, "ERROR: ");
	(void) fprintf(stderr, s, name);
	(void) fprintf(stderr, "\n");
}

void *
xmalloc(size_t len)
{
	void *ret = malloc(len);
	if (ret == 0)
		vsxerror("malloc() failed\n");
	return ret;
}

void *
xrealloc(void *ptr, size_t len)
{
	void *ret = realloc(ptr, len);
	if (ret == 0)
		vsxerror("realloc() failed\n");
	return ret;
}

/* ARGSUSED */
void
emit_line_note(char *dummy1, int dummy2)
{
	/* do nothing */
}

/* ARGSUSED */
/* VARARGS0 */
void
warning(char *dummy, ...)
{
	/* do nothing */
}

/* ARGSUSED */
void
warn_decl(tree dummy1, char *dummy2)
{
	/* do nothing */
}
